import os

fuck = input('Service name: ')

os.system(f'systemctl stop {fuck}')
os.system(f'systemctl disable {fuck}')

os.system(f'rm /etc/systemd/system/{fuck}')
# and symlinks that might be related
os.system(f'rm /etc/systemd/system/{fuck}')

os.system(f'rm /usr/lib/systemd/system/{fuck}')
# and symlinks that might be related
os.system(f'rm /usr/lib/systemd/system/{fuck}')

os.system('systemctl daemon-reload')
os.system('systemctl reset-failed')


from jag_util import dict_pretty_print


_room_echo = '[Request Evaluator]'
_rebind = print

def conlog(*args):
	print(_room_echo, *args)



def _default_room(request, response, services):
	conlog('Executing default action', request.abspath)

	# for this to work it has to be a GET request
	if request.method == 'get':

		# anything that is not inside the server shall get rejected
		if not request.abspath.resolve().is_relative_to(request.srv_res.doc_root):
			request.reject()
			return

		# first check if path explicitly points to a file
		if request.abspath.is_file():
			services.serve_file()
			return

		# if it's not a file - check whether the target dir has an index.html file
		if (request.abspath / 'index.html').is_file():
			services.serve_dir_index()
			return

		# if it's just a directory - list it
		if request.abspath.is_dir() and request.srv_res.cfg['dir_listing']['enabled']:
			services.list_dir()
			return


	# otherwise - reject
	request.reject()



# A bunch of default services the server can provide
class sv_services:
	"""
	A bunch of default services the server can provide.
	Most notable one: Serving GET requests
	"""
	def __init__(self, request, response):
		self.request = request
		self.response = response

	# Serve a file to the client in a CDN manner
	# If no file is provided - serve path from the request
	def serve_file(self, tgt_file=None, respect_range=True):
		request = self.request
		response = self.response

		if not request.abspath.is_file():
			self.request.reject()
			return

		# all good - set content type and send the shit
		response.content_type = (
			request.srv_res.mimes['signed'].get(request.abspath.suffix)
			or
			'application/octet-stream'
		)

		# if the size is too big for a single flush - stream in chunks
		# This is very important, because serving a 2kb svg in chunks slows the response time
		# and serving an 18gb .mkv Blu-ray remux in a single flush is impossible
		if request.abspath.stat().st_size > request.srv_res.cfg['static_cdn']['max_buffered_size']:
			with open(str(request.abspath), 'r+b') as f:
				# if request comes with a Range header - try serving the requested byterange
				# '0-' VERY funny, fuck right off
				if request.byterange and (request.headers['range'].strip() != '0-') and respect_range:
					response.serve_range(f)
				else:
					response.stream_buffer(f, (1024*1024)*5)
		else:
			response.flush_buffer(request.abspath.read_bytes())

		self.request.terminate()

	# List directory as an html page
	def list_dir(self, respect_htaccess=True):
		self.response.content_type = 'text/html'

		with self.response.stream_bytechunks() as stream:
			for chunk in self.request.srv_res.list_dir.dir_as_html(self.request.abspath):
				stream.send(chunk)

		self.request.terminate()

	# Serve "index.html" from the requested path
	# according to server config
	def serve_dir_index(self):
		if not (self.request.abspath / 'index.html').is_file():
			self.request.reject()

		self.response.content_type = 'text/html'
		self.response.flush_buffer(
			(self.request.abspath / 'index.html').read_bytes()
		)

	# Because why not
	def default(self):
		_default_room(self.request, self.response, self)







# Stream chunks
class _chunkstream:
	def __init__(self, cl_con):
		self.cl_con = cl_con

	def __enter__(self):
		return self

	def __exit__(self, type, value, traceback):
		self.cl_con.sendall(b'0\r\n\r\n')
		# No termination, because it's speculated,
		# that it's possible to send some sort of trailing headers

	def send(self, data):
		# send the chunk size
		self.cl_con.sendall(f"""{hex(len(data)).lstrip('0x')}\r\n""".encode())
		# send the chunk itself
		self.cl_con.sendall(data)
		# send separator
		self.cl_con.sendall(b'\r\n')




# Read part of a buffer in chunks (start:end)
# todo: There are 0 validations
# (neither end or start can be negative)
# (end cannot be smaller than start)
# (start and end cannot be the same)
# (start and end should not result into 0 bytes read)

# src  ------------------------
# rng        ^         ^
# prog       -----
# want            ---------
# let             -----
class aligned_buf_read:
	def __init__(self, buf, start, end):
		# START is inclusive
		# END is NOT inclusive
		self.buf = buf
		self.start = start
		self.end = end
		self.target_amount = end - start
		self.progress = 0

		# seek to the beginning
		self.buf.seek(start, 0)

	def read(self, amt):
		# max(smallest, min(n, largest))
		# Todo: is this slow ?
		allowed_amount = max(0, min(amt, self.end - self.progress))
		chunk = self.buf.read(allowed_amount)
		self.progress += allowed_amount
		return chunk




class sv_response:
	def __init__(self, request, cl_con, srv_res):
		self.request = request
		self.cl_con = cl_con
		self.srv_res = srv_res
		self.headers = {
			'Server': 'Jag',
		}

		self.content_type = 'application/octet-stream'
		self.code = 200

		self.offered_services = sv_services(self.request, self)

	# dump headers and response code to the client
	def send_preflight(self):
		"""
		Dump headers and response code to the client
		"""

		# send response code
		self.cl_con.sendall(
			f"""HTTP/1.1 {self.srv_res.response_codes[self.code]}\r\n""".encode()
		)

		# important todo: better way of achieving this
		self.headers['Content-Type'] = self.content_type

		# send headers
		for header_name, header_value in self.headers.items():
			self.cl_con.sendall(
				f"""{header_name}: {header_value}\r\n""".encode()
			)

		# Send an extra \r\n to indicate the end of headers
		self.cl_con.sendall('\r\n'.encode())

		# important todo: There's a built-in way to make functions only fire once
		# Yes, BUT, it costs A LOT of time and effort for the machine
		# Such a simple buttplug is WAY more efficient
		self.send_preflight = lambda: None

	# mark response as a download
	def mark_as_xfiles(self, filename):
		"""
		This is needed if you want the response body to be treated
		as a file download by the client.
		Useful when a media file, like .mp4 video should be downloaded
		by the client instead of playing back.
		"""
		self.headers['Content-Disposition'] = f'attachment; filename="{str(filename)}"'

	# - Send headers to the client
	# - Send the entirety of the provided buffer in one go
	# - Collapse connection
	def flush_buffer(self, data):
		if not isinstance(data, bytes):
			data = data.getvalue()

		# important todo: the response should either be chunked or have Content-Length header
		self.headers['Content-Length'] = len(data)

		# send headers
		self.send_preflight()

		# send the body
		self.cl_con.sendall(data)

		# terminate
		self.request.terminate()

	def stream_bytechunks(self):
		# This cannot be otherwise
		self.headers['Transfer-Encoding'] = 'chunked'
		# It's impossible to stream multiple groups of chunks
		self.send_preflight()

		return _chunkstream(self.cl_con)

	# Automated action:
	# - Add 'Transfer-Encoding: chunked' header
	# - Send headers to the client
	# - Stream the entirety of the provided buffer in chunks
	# - Collapse the connection
	def stream_buffer(self, data, chunk_size=65535):
		"""
		Send the data in chunks.
		data = io.BytesIO object.
		chunk_size = the size of a single chunk in bytes
		"""

		# The response is EITHER chunked OR has Content-Length
		self.headers['Transfer-Encoding'] = 'chunked'

		# first - dump headers
		self.send_preflight()

		# Safety (tin foil hat): Move the carret to the very beginning of the buffer
		data.seek(0, 0)
		# stream chunks
		with self.stream_bytechunks() as stream:
			while True:
				# read chunk
				chunk = data.read(chunk_size)
				# check if there's still any data
				if not chunk:
					break
				stream.send(chunk)

		# collapse the connection
		self.request.terminate()

	# Serve specified buffer according to the Range header
	# important todo: This is 100% raw/bare
	# nothing is checked or validated
	def serve_range(self, buf):
		# Set code to partial-content
		self.code = 206
		# It'd be stupid not to do it this way...
		self.headers['Transfer-Encoding'] = 'chunked'
		self.send_preflight()

		buf_size = buf.seek(0, 2)

		# begin streaming
		with self.stream_bytechunks() as stream:
			# stream all chunk groups
			# (order is preserved)
			for chunk_start, chunk_end in self.request.byterange:
				# Python is amazing: array[37:None] is a valid syntax
				_start = chunk_start
				_end = chunk_end or buf_size

				# If only end is specified - stream suffix
				# Todo: current implementation requires calculating
				# start offset, which means that buffer should be of known size
				if chunk_end and not chunk_start:
					_end = buf_size
					_start = _end - chunk_end

				aligned_reader = aligned_buf_read(buf, _start, _end)
				while True:
					data = aligned_reader.read(self.srv_res.cfg['chunk_stream_piece_size'])
					if not data:
						break
					stream.send(data)

		self.request.terminate()





class cl_request:
	def __init__(self, cl_con, cl_addr, srv_res):
		self.cl_con = cl_con
		self.cl_addr = cl_addr
		self.srv_res = srv_res

		self.headers = {}

		# Initialize the response class
		# Early init of this class is needed
		# for rejecting certain requests
		self.response = sv_response(self, cl_con, srv_res)

		# Some widely-used headers are lazily processed
		# for easier use
		self._cookie = None
		self._cache_control = None
		self._accept = None
		self._byterange = None

		# Try/Except in case of malformed request
		# Why bother?
		# It's client's responsibility to perform good requests
		try:
			self._eval_request()
		except Exception as e:
			self.reject(400)
			raise e
		

	# Init
	# =================

	# Keep eating bytes from the client
	# until the entire Request Header arrives
	def collect_head_buf(self):

		io = self.srv_res.pylib.io

		self.head_buf = io.BytesIO()
		self.body_buf = io.BytesIO()

		double_rn = 0
		expect_r = False

		# important todo: This is extremely (relatively) slow
		# simple http server from python base library does something like
		# do 1 byte receive from client till \r\n\r\n
		# OR
		# Read 65535 bytes and then process the thing
		while True:
			data = self.cl_con.recv(65535)
			for idx, char in enumerate(data):
				# conlog('Char:', chr(char).encode())

				if char != 10 and char != 13:
					# conlog('^ no match, abort')
					expect_r = False
					double_rn = 0
					continue

				if char == 13:
					# conlog('^ found 13')
					expect_r = True
					continue

				if expect_r and char != 10:
					# conlog('^ is not 10, abort')
					expect_r = False
					double_rn = 0
					continue

				if expect_r and char == 10:
					# conlog('^ found 10, +1')
					double_rn += 1
					expect_r = False

				if double_rn == 2:
					# conlog('Writing to body buffer:', bytes(data[(idx+1):]))
					self.body_buf.write(bytes(data[(idx+1):]))
					self.head_buf.write(bytes(data[:idx]))
					break

			if double_rn == 2:
				conlog('Header Buf', self.head_buf.getvalue().decode())
				# conlog('Body Buf', self.body_buf.getvalue())
				break

			self.head_buf.write(data)

	# Request evaluation has a dedicated function for easier error handling
	def _eval_request(self):
		# io = self.srv_res.pylib.io
		# sys = self.srv_res.pylib.sys
		urllib = self.srv_res.pylib.urllib
		Path = self.srv_res.pylib.Path

		# Fully custom method of receiving the Request Header
		# gives a lot of benefits
		self.collect_head_buf()

		# raw bytes of the header
		header_data = self.head_buf.getvalue()
		conlog(header_data.decode())

		# split header into lines
		header_data = header_data.decode().split('\r\n')
		conlog('\n'.join(header_data))

		# First line of the header is always the request method, path and http version
		# It's up to the client to send valid data
		self.method, self.path, self.protocol = header_data[0].split(' ')
		conlog(self.method, self.path, self.protocol)
		self.method = self.method.lower()

		# deconstruct the url into components
		parsed_url = urllib.parse.urlparse(self.path)

		# important todo: lazy processing
		# first - evaluate query params
		self.query_params = {k:(''.join(v)) for (k,v) in urllib.parse.parse_qs(parsed_url.query, True).items()}

		# then, evaluate path
		decoded_url_path = urllib.parse.unquote(parsed_url.path)
		self.abspath = self.srv_res.doc_root / Path(decoded_url_path.lstrip('/'))
		self.relpath = Path(decoded_url_path.lstrip('/'))
		self.trimpath = self.relpath

		# Delete the first line as it's no longer needed
		del header_data[0]

		# parse the remaining headers into a dict
		request_dict = {}
		for line in header_data:
			# skip empty stuff
			if line.strip() == '':
				continue
			line_split = line.split(': ')
			request_dict[line_split[0].lower()] = ': '.join(line_split[1:])

		dict_pretty_print(request_dict)

		self.headers = request_dict


	# Actions
	# =================

	# Properly collapse the tunnel between server and client
	def terminate(self):
		socket = self.srv_res.pylib.socket
		self.cl_con.shutdown(socket.SHUT_RDWR)
		self.cl_con.close()

		# Termination is only possible once
		self.terminate = lambda: None

	# Send a very simple html document
	# with a short description of the provided Status Code
	def reject(self, code=401, hint=''):
		self.response.code = code
		self.response.content_type = 'text/html'
		self.response.flush_buffer(
			self.srv_res.reject_precache
			.replace(b'$$reason', self.srv_res.response_codes[code].encode())
			.replace(b'$$hint', str(hint).encode())
		)

	# I cannot be bothered. Here, have *args and fuckoff
	def match_path(self, action_dict, *args, trim_path=True):
		"""
		Sample set/list:
		{
			('/pootis/sandwich/dispenser', func_name1),
			('/pootis',                    func_name2),
		}
		"""
		comparator = '/' + self.relpath.as_posix()

		for rpath, func in action_dict:
			# print(rpath, 'startswith', )
			if comparator.startswith(rpath):
				if trim_path:
					self.trimpath = self.srv_res.pylib.Path(comparator.lstrip(rpath))
				return func(self, self.response, self.response.offered_services, *args)

		# if no match was found - return false
		return False


	def read_body_stream(self):
		content_length = int(self.headers['content-length'])
		read_progress = self.body_buf.seek(0, 2)
		yield self.body_buf.getvalue()
		while True:
			if read_progress >= content_length:
				break
			# todo: finetune this value
			# or expose it in the config
			received_data = self.cl_con.recv(65535)
			yield received_data
			read_progress += len(received_data)


	# Utility
	# =================
	def parse_kv(self, kvs, separator=',', key_to_lower=True):
		pairs = kvs.split(separator)
		result = {}
		for pair in pairs:
			pair_split = pair.split('=')
			if key_to_lower:
				pair_split[0] = pair_split[0].upper()

			if len(pair_split) == 1:
				result[pair_split[0]] = True
				continue

			val = '='.join(pair_split[1:])
			# important todo: is this really needed ?
			try:
				val = float(val)
			# important todo: generic exceptions are very bad
			except:
				pass
			try:
				val = int(val)
			except:
				pass

			result[pair_split[0]] = val

		return result


	# Processed headers
	# =================

	@property
	def cookie(self):
		if self._cookie:
			return self._cookie

		cookie_data = self.headers.get('cookie')
		if not cookie_data:
			return None

		self._cookie = self.parse_kv(cookie_data, separator=';')

		return self._cookie

	# Even though server doesn't support advanced caching...
	@property
	def cache_control(self):
		if self._cache_control:
			return self._cache_control

		cache_data = self.headers.get('cache-control')
		if not cache_data:
			return None

		self._cache_control = self.parse_kv(cookie_data, separator=',')

		return self._cache_control


	@property
	def byterange(self):
		if self._byterange:
			return self._byterange

		range_data = self.headers.get('range')
		if not range_data:
			return None

		# todo: it's always assumed that range is in bytes
		ranges = range_data.split('=')[1].split(',')

		self._byterange = []
		for chunk in ranges:
			chunk_split = chunk.strip().split('-')
			rstart = int(chunk_split[0]) if chunk_split[0] else None
			rend =   int(chunk_split[1]) if chunk_split[1] else None
			self._byterange.append(
				(rstart, rend)
			)

		return self._byterange






# The server creates "rooms" for every incoming connection.
# The Base Room does some setup, like evaluating the request.
# Further actions depend on the server setup:
# If callback function is specified, then it's triggered
# without any automatic actions
# If callback function is NOT specified, then server provides
# Some of its default services
def base_room(cl_con, cl_addr, srv_res):
	import sys, traceback
	import importlib.util

	try:
		# precache some commonly-used python libraries
		# important todo: is this even needed?
		srv_res.reload_libs()

		# Evaluate the request
		evaluated_request = cl_request(cl_con, cl_addr, srv_res)
		conlog('Initialized basic room, evaluated request')

		# Create service object
		# offered_services = sv_services(evaluated_request, evaluated_request.response)

		# Now either pass the control to the room specified in the config
		# or the default room
		if srv_res.cfg['room_file']:
			spec = importlib.util.spec_from_file_location('main', str(srv_res.cfg['room_file']))
			custom_func = importlib.util.module_from_spec(spec)
			spec.loader.exec_module(custom_func)
			custom_func.main(
				evaluated_request,
				evaluated_request.response,
				evaluated_request.response.offered_services
			)
		else:
			_default_room(
				evaluated_request,
				evaluated_request.response,
				evaluated_request.response.offered_services
			)

	except Exception as err:
		conlog(
			''.join(
				traceback.format_exception(
					type(err),
					err,
					err.__traceback__
				)
			)
		)

		_trback = ''.join(
			traceback.format_exception(
				type(err),
				err,
				err.__traceback__
			)
		)
		cl_con.sendall('HTTP/1.1 500 Internal Server Error\r\n'.encode())
		_rcontent = f"""<!DOCTYPE HTML>
			<html>
				<head>
					<meta http-equiv="Content-Type" content="text/html;charset=utf-8">
					<title>Rejected</title>
				</head>
				<body>
					<h1>500 Internal Server Error</h1>
					<h3>Server: Jag</h3>
					<p style="white-space: pre;">{_trback}</p>
				</body>
			</html>
		"""
		cl_con.sendall(f'Content-Length: {len(_rcontent)}\r\n\r\n'.encode())
		cl_con.sendall(_rcontent.encode())

		raise err


	cl_con.close()
	sys.exit()








def dict_pretty_print(d):
	sex = '\n'
	for key in d:
		sex += f"""{('>' + str(key) + '<').ljust(30)} :: >{str(d[key])}<""" + '\n'

	print(sex)

def multireplace(src, replace_pairs):
	for replace_what, replace_with in replace_pairs:
		src = src.replace(replace_what, replace_with)
	return src


def clamp(num, tgt_min, tgt_max):
	return max(tgt_min, min(num, tgt_max))


import socket, threading, time, sys, hashlib, json, base64, struct, io, multiprocessing
from pathlib import Path
import traceback

# important todo: wat ?
# (this library simply has to be a proper package)
sys.path.append(str(Path(__file__).parent))

from base_room import base_room
import jag_util

_main_init = '[root]'
_server_proc = '[Server Process]'




# Path
# jag_util
# socket
# threading
# time
# sys
# hashlib
# json
# base64
# struct
# io
# multiprocessing
class pylib_preload:
	"""precache python libraries"""
	def __init__(self):
		import socket
		import threading
		import time
		import sys
		import hashlib
		import json
		import base64
		import struct
		import io
		import multiprocessing
		import traceback
		import urllib
		import math

		from pathlib import Path

		import jag_util

		self.jag_util =  jag_util

		self.Path =      Path
		self.socket =    socket
		self.threading = threading
		self.time =      time
		self.sys =       sys
		self.hashlib =   hashlib
		self.json =      json
		self.base64 =    base64
		self.struct =    struct
		self.io =        io
		self.traceback = traceback
		self.urllib =    urllib
		self.math =      math




# sysroot         = Path-like pointing to the root of the jag package
# pylib           = A bunch of precached python packages
# mimes           = A dictionary of mime types; {file_ext:mime}
#                   | regular = {file_ext:mime}
#                   | signed =  {.file_ext:mime}
# response_codes  = HTTP response codes {code(int):string_descriptor}
# reject_precache = HTML document which sez "access denied"
# cfg             = Server Config
# doc_root        = Server Document Root
# list_dir        = List directory as html document
class server_info:
	"""
	Server info.
	This class contains the config itself,
	some preloaded python libraries,
	and other stuff
	"""
	def __init__(self, config=None):
		from mimes.mime_types_base import base_mimes
		from mimes.mime_types_base import base_mimes_signed
		from response_codes import codes as _rcodes

		from pathlib import Path
		import io
		import jag_util

		config = config or {}

		# root of the python package
		self.sysroot = Path(__file__).parent

		# extend python paths with included libs
		sys.path.append(str(self.sysroot / 'libs'))

		# mimes
		self.mimes = {
			'regular': base_mimes,
			'signed': base_mimes_signed,
		}

		# HTTP response codes
		self.response_codes = _rcodes

		# Reject document precache
		self.reject_precache = (self.sysroot / 'assets' / 'reject.html').read_bytes()

		# Base config
		self.cfg = {
			# Port to run the server on
			'port': 0,

			# Document root (where index.html is)
			'doc_root': None,

			# This path should point to a python file with "main()" function inside
			# If nothing is specified, then default room is created
			'room_file': None,

			# Could possibly be treated as bootleg anti-ddos/spam
			'max_connections': 0,

			# The name of the html file to serve when request path is '/'
			'root_index': None,
			'enable_indexes': True,
			'index_names': ['index.html'],

			# The size of a single chunk when streaming content in HTTP chunks
			'chunk_stream_piece_size': (1024*1024)*10,
		} | config

		self.doc_root = Path(self.cfg['doc_root'])

		# Directory Listing
		self.cfg['dir_listing'] = {
			'enabled': True,
			'dark_theme': False,
		} | (config.get('dir_listing') or {})

		if self.cfg['dir_listing']['enabled']:
			from dir_list import dirlist
			self.list_dir = dirlist(self)

		# Advanced CDN serving
		self.cfg['static_cdn'] = {
			'max_buffered_size': (1024*1024)*10,
			# Path to the static CDN
			# can point anywhere
			'path': None,
			# Relieve the filesystem stress by precaching items inside this folder
			# only useful if folder contains a big amount of small files
			'precache': True,
			# An array of paths relative to the root cdn path
			# to exclude from caching
			'cache_exclude': [],
			# Glob pattern for caching files, default to '*'
			'pattern': None,
			# Wether to trigger the callback function
			# when incoming request is trying to access the static CDN
			'skip_callback': True,
		} | (config.get('static_cdn') or {})

		self.cdn_path = None
		self.cdn_cache = {}

		if self.cfg['static_cdn']['path']:
			self.cdn_path = Path(self.cfg['static_cdn']['path'])
			self.precache_cdn()

	def precache_cdn(self):
		for file in self.cdn_path.rglob(self.cfg['static_cdn']['pattern'] or '*'):
			self.cdn_cache[file.relative_to(self.cdn_path).as_posix()] = \
				io.BytesIO(file.read_bytes())

		print(
			_server_proc,
			'precached CDN',
			jag_util.dict_pretty_print(self.cdn_cache)
		)

	def reload_libs(self):
		import time
		ded = time.time()
		# preload python libraries
		self.pylib = pylib_preload()
		print('Preloaded libs in', (time.time() - ded)*1000)



def sock_server(sv_cfg):
	# Preload resources n stuff
	print(_server_proc, 'Preloading resources...')
	server_resources = server_info(sv_cfg)
	print(_server_proc, 'Binding server to a port...')
	# Port to run the server on
	# port = 56817
	port = server_resources.cfg['port']
	# Create the Server object
	s = socket.socket()
	# Bind server to the specified port. 0 = Find the closest free port and run stuff on it
	s.bind(
		('192.168.0.10', port)
	)

	# Basically launch the server
	# The number passed to this function identifies the max amount of simultaneous connections
	# If the amount of connections exceeds this limit -
	# connections become rejected till other ones are resolved (aka closed)
	# 0 = infinite
	s.listen(server_resources.cfg['max_connections'])

	print(_server_proc, 'Server listening on port', s.getsockname()[1])

	# important todo: does this actually slow the shit down?
	# important todo: is it just me or this crashes the system ???!!?!??!?!?!?!?
	# important todo: this creates a bunch of threads as a side effect
	# important todo: Pickling is EXTREMELY slow and bad

	# Multiprocess pool automatically takes care of a bunch of stuff
	# But most importantly, it takes care of shadow processess left after collapsed rooms
	# (linux moment)
	with multiprocessing.Pool() as pool:
		while True:
			print(_server_proc, 'Entering the main listen cycle which would spawn rooms upon incoming connection requests...')
			# Try establishing connection, nothing below this line gets executed
			# until server receives a new connection
			conn, address = s.accept()
			print(_server_proc, 'Got connection, spawning a room. Client info:', address)
			# Create a basic room
			pool.apply_async(base_room, (conn, address, server_resources))

			print(_server_proc, 'Spawned a room, continue accepting new connections')


def server_process(srv_params, stfu=False):
	print(_main_init, 'Creating and starting the server process...')
	# Create a new process containing the main incoming connections listener
	server_ctrl = multiprocessing.Process(target=sock_server, args=(srv_params,))
	print(_main_init, 'Created the process instructions, attempting launch...')
	# Initialize the created process
	# (It's not requred to create a new variable, it could be done in 1 line with .start() in the end)
	server_ctrl.start()

	print(_main_init, 'Launched the server process...')




if __name__ == '__main__':
	server_params = {
		'doc_root': r'E:\!webdesign\jag',
		'port': 56817,
		'dir_listing': {
			'enabled': True,
		}
	}
	server_process(server_params)









import os, socket, shutil
from pathlib import Path

def clear():
	os.system('cls' if os.name == 'nt' else 'clear')

def is_port_in_use(port):
	with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
		return s.connect_ex(('localhost', port)) != 0

def dict_replace(src, matches):
	for drp in matches:
		src = src.replace(drp, matches[drp])
	return src

thisdir = Path(__file__).parent

class ez_service:
	def __init__(self):
		self.prmdict = {
			'docroot': None,
			'service_name': None,
			'run_user': None,
			'port': None,
			'pypath': None,
			'cfgloc': None,
		}

		self.msg = ''

		clear()
		self.mainl()

	def print_stats(self):
		clear()
		print('')
		for idx, stat in enumerate(self.prmdict):
			if self.prmdict[stat]:
				print(f'{idx+1} - {stat.ljust(20)}:', self.prmdict[stat])

		print(self.msg)


	def mainl(self):
		self._docroot()
		self.msg = ''
		self._service_name()
		self.msg = ''
		self._run_user()
		self.msg = ''
		self._port()
		self.msg = ''
		self._pypath()
		self.msg = ''
		self._cfgloc()
		self.msg = ''

		while True:
			self.print_stats()
			print('Is the above info correct?')
			print('If you want to tweak something - press a number')
			print('If you are happy - type "y"')
			decide = input('Your answer: ')
			if decide == 'y':
				break

			decide_dict = {
				'1': self._docroot,
				'2': self._service_name,
				'3': self._run_user,
				'4': self._port,
				'5': self._cfgloc,
			}

			if not decide in decide_dict:
				continue

			decide_dict[decide]()



	def _service_name(self):
		while True:
			service_name = input('Enter service name: ')
			if not service_name.strip():
				self.print_stats()
				self.msg = 'Invalid service name'
				continue
			break

		self.prmdict['service_name'] = service_name

	def _run_user(self):
		while True:
			self.print_stats()
			run_user = input('User to run as ("-" for www-data): ')
			if not run_user.strip():
				self.msg = 'Invalid username'
				continue

			if run_user.strip() == '-':
				run_user = 'www-data'
				break
			break

		self.prmdict['run_user'] = run_user

	def _port(self):
		while True:
			self.print_stats()
			port = input('Port: ')
			try:
				port = int(port)
				is_port_in_use(port)
			except Exception as e:
				self.msg = 'Invalid port'
				continue
			break

		self.prmdict['port'] = port

	def _docroot(self):
		while True:
			self.print_stats()
			docroot = input('Document Root: ')
			if not Path(docroot.strip('"')).is_dir() or not docroot.strip('"'):
				self.msg = 'Invalid directory'
				continue
			break

		self.prmdict['docroot'] = docroot.strip()

	def _cfgloc(self):
		while True:
			self.print_stats()
			cfgloc = input('Config Location (create new folder named after the service there): ')
			if not Path(cfgloc.strip('"')).is_dir() or not cfgloc.strip('"'):
				self.msg = 'Invalid directory'
				continue
			break

		self.prmdict['cfgloc'] = cfgloc.strip()

	def _pypath(self):
		while True:
			self.print_stats()
			pypath = input('Path to python executable (/usr/bin/python3): ')
			if not Path(pypath.strip('"')).is_file(): 
				self.msg = 'Invalid path'
				continue
			break

		self.prmdict['pypath'] = pypath.strip()




if __name__ == '__main__':
	fuckoff = ez_service()

	unit_match = {
		'$JAG_SERVICENAME$': fuckoff.prmdict['service_name'],
		'$JAG_CFGLOC$': str(Path(fuckoff.prmdict['cfgloc']) / fuckoff.prmdict['service_name']).rstrip('/'),
	}
	unit = dict_replace((thisdir / 'templates' / 'unit.service').read_text(), unit_match)

	Path(f"""/etc/systemd/system/{fuckoff.prmdict['service_name']}.service""").write_text(unit)

	cfgdir = Path(fuckoff.prmdict['cfgloc'])
	cfgdir.mkdir(exist_ok=True)
	cfgdir = cfgdir / fuckoff.prmdict['service_name']
	cfgdir.mkdir(exist_ok=True)
	(cfgdir / 'conf-enabled').mkdir(exist_ok=True)

	basecfg_match = {
		'$JAG_DOCROOT$':         fuckoff.prmdict['docroot'].rstrip('/'),
		'$JAG_PIDFILE$':         str(Path('/run') / f"""{fuckoff.prmdict['service_name']}.pid""").rstrip('/'),
		'$JAG_USERNAME$':        fuckoff.prmdict['run_user'],
		'$JAG_PORTNUM$':         str(fuckoff.prmdict['port']),
		'$JAG_INCLUDE_CFG_DIR$': str(Path(cfgdir) / 'conf-enabled').rstrip('/'),
	}
	basecfg = dict_replace((thisdir / 'templates' / 'light_httpd' / 'server_prms.conf').read_text(), basecfg_match)
	(cfgdir / 'server_prms.conf').write_text(basecfg)

	cgicfg_match = {
		'$JAG_PYPATH$': fuckoff.prmdict['pypath'].rstrip('/'),
	}
	basecfg = dict_replace((thisdir / 'templates' / 'light_httpd' / 'conf-enabled' / '10-cgi.conf').read_text(), cgicfg_match)
	(cfgdir / 'conf-enabled' / '10-cgi.conf').write_text(basecfg)

	shutil.copy(
		thisdir / 'templates' / 'light_httpd' / 'conf-enabled' / '10-status.conf',
		cfgdir / 'conf-enabled' / '10-status.conf'
	)
	shutil.copy(
		thisdir / 'templates' / 'light_httpd' / 'conf-enabled' / '10-accesslog.conf',
		cfgdir / 'conf-enabled' / '10-accesslog.conf'
	)

	os.system('systemctl daemon-reload')
	os.system(f"""systemctl enable {fuckoff.prmdict['service_name']}""")
	os.system(f"""service {fuckoff.prmdict['service_name']} start""")
	print('')
	print('Done')
	print('')







class dirlist:
	"""simple directory lister"""
	def __init__(self, srv_res):
		"""
		Init compiles a very simple document for later use
		"""
		from bs4 import BeautifulSoup
		import base64
		import jag_util

		self.srv_res = srv_res

		# shortcut to the dir_listing folder in the assets
		assets_dir = srv_res.sysroot / 'assets' / 'dir_listing'

		# Parse the template to populate it with scripts and styles
		doc_base = BeautifulSoup(
			(assets_dir / 'base.html').read_bytes(),
			'html.parser'
		)

		# Load css file and add icons to it
		css_base = jag_util.multireplace(
			(assets_dir / 'style.css').read_bytes(),
			[
				(b'$$folder_icon_b64', base64.b64encode((assets_dir / 'dir_icon.svg').read_bytes())),
				(b'$$file_icon_b64', base64.b64encode((assets_dir / 'file_icon.svg').read_bytes())),
				(b'$$dl_icon_b64', base64.b64encode((assets_dir / 'download_icon.svg').read_bytes())),
			]
		)
		# Append css to the body
		style_tag = doc_base.new_tag('style', type='text/css')
		style_tag.string = css_base.decode()
		doc_base.body.append(style_tag)

		# add script to the file
		script_tag = doc_base.new_tag('script', type='text/javascript')
		script_tag.string = (assets_dir / 'script.js').read_text()
		doc_base.body.append(script_tag)

		# split the document into 2 parts
		self.doc_a, self.doc_b = str(doc_base).encode().split(b'$$speed_split')


	def dir_as_html(self, tgt_dir):
		# import io
		# from pathlib import Path
		io = self.srv_res.pylib.io
		Path = self.srv_res.pylib.Path

		# dir to list
		tgt_dir = Path(tgt_dir)

		# yield the beginning of the document
		# important todo: there's certainly a better way of doing this,
		# like having predefined buffers or something
		yield self.doc_a

		# raw glob result
		dirlist = list(tgt_dir.glob('*'))
		# sort by name
		dirlist.sort(key=lambda a: a.name.lower())
		# sort by type (folder/file)
		dirlist.sort(key=lambda a: int(a.is_file()))


		# yield listed entries one by one
		# in a form of html
		for entry in dirlist:
			yield (
				f"""
					<div class="list_entry" {'file_entry' if entry.is_file() else 'dir_entry'}>
						<a
							class="actual_link"
							href="/{entry.relative_to(self.srv_res.doc_root).as_posix()}"
						>{entry.name}</a>
						<a download class="dl_button" href="/{entry.relative_to(self.srv_res.doc_root).as_posix()}"></a>
					</div>
				""".encode()
			)

		# yield the end of the document
		yield self.doc_b

		# return list_buf











class jag_server:
	def __init__(self):
		import cgi, sys, cgitb
		import os, json, platform
		from pathlib import Path

		self._fatal_err_msg =         """$JAG_FATAL_ERR_MSG$"""
		self._regular_err_msg =       """$JAG_REGULAR_ERR_MSG$"""
		self._jtw_no_action_reason =  """$JAG_JATEWAY_ACTION_NOT_FOUND_MSG$"""
		self._xfiles_hname =          """$JAG_X_FILES_HNAME$"""
		self._action_url_param_name = """$JAG_ACT_URL_PRM_NAME$"""

		self.response_content_type = 'application/octet-stream'


		# traceback messages
		cgitb.enable(format='text')

		# don't import modules many times...
		self.cgitb = cgitb
		self.Path = Path
		self.json = json
		self.sys = sys
		self.os = os

		self.output = sys.stdout.buffer.write

		# buffer to flush
		self.sv_buffer = b''

		# parse url params into a dict, if any
		get_cgi_params = cgi.parse()
		url_params = {}
		for it in get_cgi_params:
			url_params[it] = ''.join(get_cgi_params[it])
		self.prms = url_params

		# parse http headers into a dict, if any
		self.headers = {}
		for hd in os.environ:
			# print(hd, os.environ[hd], '\n')
			if hd.startswith('HTTP_'):
				self.headers[hd.replace('HTTP_', '').lower()] = os.environ[hd]

		# get cookies
		# important todo: is it really neccessary to postpone cookie evaluation ?
		self._cookies = None
		self._outp_cookies = {}


		# read body content, if any
		self.bin = b''
		# response headers are the ones this script is about to output to the client
		self.response_headers = {}
		# todo: there are definitely better ways of determining whether the body content could be read or not
		try:
			self.bin = sys.stdin.buffer.read()
		except:
			pass

		# todo: just don't bother and distribute two separate versions for linux and for windows ?
		self.platform = platform.system().lower()

		# also make cgitb write errors to files
		# if $JAG_BURN
		# cgitb.enable(format='text', logdir=str(self.sysdb_path / 'cgi_err'))







	def bin_as_json(self):
		return self.json.loads(self.bin)


	@property
	def cookies(self):
		if self._cookies:
			return self._cookies

		for cookie in self.os.environ['HTTP_COOKIE'].split(';'):
			csplit = cookie.split('=')
			self._cookies[csplit[0]] = '='.join(csplit[1:])


	def set_cookie(self, cname, cval):
		self._outp_cookies[cname] = cval


	# @property
	# def tr_type(self):
	# 	return self._tr_type

	# @tr_type.setter
	# def tr_type(self, newname):
		# pass

	# add error header with specified data
	def fatal_error(self, err):
		self.set_header(self._fatal_err_msg, str(err))

	def error(self, err):
		self.set_header(self._regular_err_msg, str(err))


	# spit shit
	# either fill the buffer gradually with .bin_write and then flush
	# or flush bytes immediately by passing bytes to the flush function
	def flush(self, add_b=None):
		if add_b:
			self.bin_write(add_b)
		# add headers
		for h in self.response_headers:
			# try:
			# 	hv = self.response_headers[h].encode()
			# except:
			# 	hv = str(self.response_headers[h]).encode()

			# try:
			# 	h = h.encode()
			# except:
			# 	h = str(h).encode()

			# todo: what if either key/value has \r\n in it ?
			self.output(f'{h}: {self.response_headers[h]}\r\n'.encode())

		# add cookies
		for biscuit in self._outp_cookies:
			self.output(f"""Set-Cookie: {biscuit}={self._outp_cookies[biscuit]}\r\n""".encode())

		# content type
		self.output(f'Content-Type: {self.response_content_type}\r\n\r\n'.encode())
		# buffer
		self.output(self.sv_buffer)
		# do flush
		self.sys.stdout.buffer.flush()
		self.sys.stdout.flush()
		self.sys.exit()


	def flush_json(self, j):
		self.bin_jwrite(j)
		self.flush()


	def set_header(self, hkey, hval):
		self.response_headers[hkey] = hval

	# add to bin
	# expects bytes
	def bin_write(self, dat):
		if not isinstance(dat, bytes):
			raise Exception('Jag: Cant add anything besides bytes to output buffer')

		self.sv_buffer += dat

	# add json to bin
	def bin_jwrite(self, jsn):
		self.sv_buffer += self.json.dumps(jsn).encode()


	# spit file
	def x_files(self, flpath=None, flname=None):

		if not flpath or not flname:
			raise Exception('Jag: x_files transfer: one of the arguments is completely invalid')

		floc = self.Path(flpath)

		if not floc.is_file():
			raise Exception('Jag: x_files transfer: file path does not exist')

		# src from hello.py:
		# sys.stdout.write(b'Content-Type: application/octet-stream\r\n')
		# sys.stdout.write(b'Content-Disposition: attachment; filename="bigfile.mp4"\r\n')
		# sys.stdout.write(b'X-Sendfile: /home/basket/scottish_handshake/db/20181225_182650.ts\r\n\r\n')

		self.output('Content-Type: application/octet-stream\r\n'.encode())
		self.output(f"""Content-Disposition: attachment; filename="{str(flname)}"\r\n""".encode())
		self.output(f"""{self._xfiles_hname}: {str(floc)}\r\n\r\n""".encode())

		self.sys.stdout.buffer.flush()
		self.sys.stdout.flush()
		self.sys.exit()


	# it shouldn't be here, but it's here because of performance reasons
	# loads json from a path
	def jload(self, pth):
		return self.json.loads(self.Path(pth).read_bytes())




class jateway:
	"""
	This gateway allows executing functions through a proxy
	to properly catch errors and report them to the client
	"""
	def __init__(self, srv, registry={}):
		self.reg = registry
		self.srv = srv
		self.action = self.srv.prms.get(self.srv._action_url_param_name)

		self._eval_action()

	def _eval_action(self):
		try:
			# if action from url params is in the registry then execute
			if self.action in self.reg:
				self.reg[self.action]()
			else:
				# todo: is it really a fatal error?
				self.srv.fatal_error('invalid_action')
				self.srv.flush(f"""Details: Requested: {self.action}, Available: {self.reg}""".encode())
		except Exception as e:
			# The way this works is pretty interesting:

			# When exception occurs upon module action evaluation - 
			# the server outputs a header to the output buffer which indicates that a fatal error occured
			# and then raises the error

			# When an error is raised, the cgitb module adds a PROPERLY FORMATTED error traceback to the output buffer
			# IN ADDITION to the header indicating that a fatal error occured
			# Not only that, but cgitb also dumps the error to the logs folder

			# The most important part is that all of the above happens
			# while keeping the client aware of the said fatal error

			# todo: cgitb simply binds a handler to exceptions. Do it manually without cgitb ?

			self.srv.output(f'{self.srv._fatal_err_msg}: {self.srv._jtw_no_action_reason}\r\n'.encode())
			# todo: citb adds appropriate content type automatically
			self.srv.output('Content-Type: application/octet-stream\r\n\r\n'.encode())

			raise e


	def eval_action(self):
		return True








base_mimes = {
	'bin':    'application/octet-stream',
	'mov':    'video/quicktime',
	'aac':    'audio/aac',
	'abw':    'application/x-abiword',
	'arc':    'application/x-freearc',
	'avif':   'image/avif',
	'avi':    'video/x-msvideo',
	'azw':    'application/vnd.amazon.ebook',
	'bmp':    'image/bmp',
	'bz':     'application/x-bzip',
	'bz2':    'application/x-bzip2',
	'cda':    'application/x-cdf',
	'csh':    'application/x-csh',
	'css':    'text/css',
	'csv':    'text/csv',
	'doc':    'application/msword',
	'docx':   'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
	'eot':    'application/vnd.ms-fontobject',
	'epub':   'application/epub+zip',
	'gz':     'application/gzip',
	'gif':    'image/gif',
	'htm':    'text/html',
	'html':   'text/html',
	'ico':    'image/vnd.microsoft.icon',
	'ics':    'text/calendar',
	'jar':    'application/java-archive',
	'jpeg':   'image/jpeg',
	'jfif':   'image/jpeg',
	'jpg':    'image/jpeg',
	'js':     'text/javascript',
	'json':   'application/json',
	'jsonld': 'application/ld+json',
	'mid':    'audio/midi audio/x-midi',
	'midi':   'audio/midi audio/x-midi',
	'mjs':    'text/javascript',
	'mp3':    'audio/mpeg',
	'mp4':    'video/mp4',
	'mpeg':   'video/mpeg',
	'mpkg':   'application/vnd.apple.installer+xml',
	'odp':    'application/vnd.oasis.opendocument.presentation',
	'ods':    'application/vnd.oasis.opendocument.spreadsheet',
	'odt':    'application/vnd.oasis.opendocument.text',
	'oga':    'audio/ogg',
	'ogv':    'video/ogg',
	'ogx':    'application/ogg',
	'opus':   'audio/opus',
	'otf':    'font/otf',
	'png':    'image/png',
	'apng':   'image/apng',
	'pdf':    'application/pdf',
	'php':    'application/x-httpd-php',
	'ppt':    'application/vnd.ms-powerpoint',
	'pptx':   'application/vnd.openxmlformats-officedocument.presentationml.presentation',
	'rar':    'application/vnd.rar',
	'rtf':    'application/rtf',
	'sh':     'application/x-sh',
	'svg':    'image/svg+xml',
	'tar':    'application/x-tar',
	'tif':    'image/tiff',
	'tiff':   'image/tiff',
	'ts':     'video/mp2t',
	'ttf':    'font/ttf',
	'txt':    'text/plain',
	'vsd':    'application/vnd.visio',
	'wav':    'audio/wav',
	'weba':   'audio/webm',
	'webm':   'video/webm',
	'webp':   'image/webp',
	'woff':   'font/woff',
	'woff2':  'font/woff2',
	'xhtml':  'application/xhtml+xml',
	'xls':    'application/vnd.ms-excel',
	'xlsx':   'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
	'xml':    'application/xml',
	'xul':    'application/vnd.mozilla.xul+xml',
	'zip':    'application/zip',
	'3gp':    'video/3gpp; audio/3gpp',
	'3g2':    'video/3gpp2; audio/3gpp2',
	'7z':     'application/x-7z-compressed',
}


base_mimes_signed = {('.'+key):value for (key,value) in base_mimes.items()}






codes = {
	# Information responses
	100: '100 Continue',
	101: '101 Switching Protocols',
	102: '102 Processing',
	103: '103 Early Hints',

	# Successful responses
	200: '200 OK',
	201: '201 Created',
	202: '202 Accepted',
	203: '203 Non-Authoritative Information',
	204: '204 No Content',
	205: '205 Reset Content',
	206: '206 Partial Content',
	207: '207 Multi-Status',
	208: '208 Already Reported',
	226: '226 IM Used',

	# Redirection messages
	300: '300 Multiple Choices',
	301: '301 Moved Permanently',
	302: '302 Found',
	303: '303 See Other',
	304: '304 Not Modified',
	307: '307 Temporary Redirect',
	308: '308 Permanent Redirect',

	# Client error responses
	400: '400 Bad Request',
	401: '401 Unauthorized',
	402: '402 Payment Required',
	403: '403 Forbidden',
	404: '404 Not Found',
	405: '405 Method Not Allowed',
	406: '406 Not Acceptable',
	407: '407 Proxy Authentication Required',
	408: '408 Request Timeout',
	409: '409 Conflict',
	410: '410 Gone',
	411: '411 Length Required',
	412: '412 Precondition Failed',
	413: '413 Payload Too Large',
	414: '414 URI Too Long',
	415: '415 Unsupported Media Type',
	416: '416 Range Not Satisfiable',
	417: '417 Expectation Failed',
	418: """418 I'm a teapot""",
	421: '421 Misdirected Request',
	422: '422 Unprocessable Content',
	423: '423 Locked',
	424: '424 Failed Dependency',
	425: '425 Too Early',
	426: '426 Upgrade Required',
	428: '428 Precondition Required',
	429: '429 Too Many Requests',
	431: '431 Request Header Fields Too Large',
	451: '451 Unavailable For Legal Reasons',

	# Server error responses
	500: '500 Internal Server Error',
	501: '501 Not Implemented',
	502: '502 Bad Gateway',
	503: '503 Service Unavailable',
	504: '504 Gateway Timeout',
	505: '505 HTTP Version Not Supported',
	506: '506 Variant Also Negotiates',
	507: '507 Insufficient Storage',
	508: '508 Loop Detected',
	510: '510 Not Extended',
	511: '511 Network Authentication Required',
}



